from __future__ import annotations
import os
import sys
import torch
import warnings
import logging
from typing import Optional, Callable
from datasets import load_from_disk, Dataset
from tqdm import tqdm
from transformers import (
    Wav2Vec2Processor, 
    AutoModelForCTC
)
from huggingsound.utils import get_chunks, get_waveforms, get_dataset_from_dict_list
from huggingsound.token_set import TokenSet
from huggingsound.normalizer import DefaultTextNormalizer
from huggingsound.trainer import TrainingArguments, ModelArguments, finetune_ctc
from huggingsound.speech_recognition.decoder import Decoder, GreedyDecoder
from huggingsound.metrics import cer, wer

# Setup logging
logger = logging.getLogger(__name__)
logging.basicConfig(
    format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
    datefmt="%m/%d/%Y %H:%M:%S",
    handlers=[logging.StreamHandler(sys.stdout)],
)
logger.setLevel(logging.INFO)

class SpeechRecognitionModel():
    """
    Speech Recognition Model.

    Parameters
    ----------
    model_path : str
        The path to the model or the model identifier from huggingface.co/models.
    
    device: Optional[str] = "cpu"
        Device to use for inference/evaluation/training, default is "cpu". If you want to use a GPU for that, 
        you'll probably need to specify the device as "cuda"
    """

    def __init__(self, model_path: str, device: Optional[str] = "cpu"):
        
        self.model_path = model_path
        self.device = device
        
        logger.info("Loading model...")
        self._load_model()

    @property
    def is_finetuned(self):
        return self.processor is not None

    def _load_model(self):

        self.model = AutoModelForCTC.from_pretrained(self.model_path)
        self.model.to(self.device)

        try:
            self.processor = Wav2Vec2Processor.from_pretrained(self.model_path)
            self.token_set = TokenSet.from_processor(self.processor)
        except Exception:
            logger.warning("Not fine-tuned model! You'll need to fine-tune it before use this model for audio transcription")
            self.processor = None
            self.token_set = None

    def transcribe(self, paths: list[str], batch_size: Optional[int] = 1, decoder: Optional[Decoder] = None) -> list[dict]:
        """ 
        Transcribe audio files.

        Parameters:
        ----------
            paths: list[str]
                List of paths to audio files to transcribe

            batch_size: Optional[int] = 1
                Batch size to use for inference

            decoder: Optional[Decoder] = None
                Decoder to use for transcription. If you don't specify this, the engine will use the GreedyDecoder.

        Returns:
        ----------
            list[dict]:
                A list of dictionaries containing the transcription for each audio file:

                [{
                    "transcription": str,
                    "start_timesteps": list[int],
                    "end_timesteps": list[int],
                    "probabilities": list[float]
                }, ...]
        """

        if not self.is_finetuned:
            raise ValueError("Not fine-tuned model! Please, fine-tune the model first.")
        
        if decoder is None:
            decoder = GreedyDecoder(self.token_set)

        sampling_rate = self.processor.feature_extractor.sampling_rate
        result = []

        for paths_batch in tqdm(list(get_chunks(paths, batch_size))):

            waveforms = get_waveforms(paths_batch, sampling_rate)

            inputs = self.processor(waveforms, sampling_rate=sampling_rate, return_tensors="pt", padding=True, do_normalize=True)

            with torch.no_grad():
                logits = self.model(inputs.input_values.to(self.device), attention_mask=inputs.attention_mask.to(self.device)).logits

            result += decoder(logits)

        return result 

    def evaluate(self, references: list[dict], predictions: Optional[list[dict]] = None, metrics_batch_size: Optional[int] = None, 
                 inference_batch_size: Optional[int] = 1, decoder: Optional[Decoder] = None, text_normalizer: Callable[[str], str] = None) -> dict:
        """ 
        Evaluate the model.

        Parameters:
        ----------
            references: list[dict]
                List of dictionaries containing the reference transcriptions for each audio file.
                The dictionaries should have the following structure:

                [{
                    "transcription": str,
                    "path": str,
                }, ...]

            predictions: Optional[list[dict]] = None
                List of dictionaries containing the predictions for each audio file.
                If this list is not provided, the engine will execute the transcribe() using the references.
                The dictionaries should have the following structure:

                [{
                    "transcription": str,
                }, ...]

            metrics_batch_size: Optional[int] = None
                Batch size to use for evaluation. When this value is specified, the evaluation function will chunk the data into 
                batches of the specified size and compute the metrics on each batch.
                After all batches are computed, the function will compute the average metrics over all batches.
                (You will probably need to define this if you have memory issues).

            inference_batch_size: Optional[int] = 1
                Batch size to use for inference.

            decoder: Optional[Decoder] = None
                Decoder to use for transcription. If you don't specify this, the engine will use the GreedyDecoder.
            
            text_normalizer: Callable[[str], str] = None
                Function used to normalize the transcriptions before evaluation.

        Returns:
        ----------
            dict:
                A dictionary containing the evaluation metrics:

                {
                    "cer": float,
                    "wer": float,
                }
        """

        if not self.is_finetuned:
            raise ValueError("Not fine-tuned model! Please, fine-tune the model first.")

        if decoder is None:
            decoder = GreedyDecoder(self.token_set)

        if text_normalizer is None:
            text_normalizer = DefaultTextNormalizer(self.token_set)

        if predictions is None:
            paths = [x["path"] for x in references]
            predictions = self.transcribe(paths, inference_batch_size, decoder)
        
        evaluation = {}
        reference_transcriptions = [text_normalizer(x["transcription"]) for x in references]
        predicted_transcriptions = [text_normalizer(x["transcription"]) for x in predictions]

        evaluation = {
            "cer": cer(predictions=predicted_transcriptions, references=reference_transcriptions, chunk_size=metrics_batch_size),
            "wer": wer(predictions=predicted_transcriptions, references=reference_transcriptions, chunk_size=metrics_batch_size)
        }

        return evaluation

    def _prepare_dataset_for_finetuning(self, dataset: Dataset, processor: Wav2Vec2Processor, text_normalizer: Callable[[str], str],
                                        length_column_name: str, num_workers: int) -> Dataset:
    
        def __process_dataset_sample(sample, text_normalizer=text_normalizer, 
                                    processor=processor, length_column_name=length_column_name):

            # Build input values
            sampling_rate = processor.feature_extractor.sampling_rate
            try:
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    waveform = get_waveforms([sample["path"]], sampling_rate)[0]
            except Exception as e:
                logger.error(f"Loading error for {sample['path']}: {e}")
                raise e
            sample["input_values"] = processor(waveform, sampling_rate=sampling_rate, do_normalize=True).input_values[0]

            # Building labels
            ## appending a " " because the CTC loss concatenates all batches into a single vector, so we need to separate sentences by a whitespace
            transcription = text_normalizer(sample["transcription"]) + " " 
            with processor.as_target_processor():
                sample["labels"] = processor(transcription).input_ids

            # Building length
            sample[length_column_name] = len(sample["input_values"])
            
            return sample

        dataset = dataset.map(
            __process_dataset_sample,
            remove_columns=dataset.column_names,
            num_proc=num_workers
        )

        return dataset

    def _get_dataset(self, processor: Wav2Vec2Processor, text_normalizer: Callable[[str], str], data: Optional[list[dict]] = None, 
                     data_cache_dir: Optional[str] = None, length_column_name: Optional[str] = "length", 
                     num_workers: Optional[int] = None) -> Dataset:

        assert data is not None or data_cache_dir is not None, "at least one of data parameters (data or data_cache_dir) needs to be specified"

        if data_cache_dir is not None and os.path.isfile(os.path.join(data_cache_dir, "dataset_info.json")):
            logger.info("Loading data from cache...")
            dataset = load_from_disk(data_cache_dir)
        else:
            logger.info("Converting data format...")
            dataset = get_dataset_from_dict_list(data)
            logger.info("Preparing data input and labels...")
            dataset = self._prepare_dataset_for_finetuning(dataset, processor, text_normalizer, length_column_name, num_workers)

            if data_cache_dir is not None:
                logger.info("Caching data...")
                dataset.save_to_disk(data_cache_dir)

        return dataset

    def finetune(self, output_dir: str, train_data: list[dict] = None, eval_data: Optional[list[dict]] = None, 
                 data_cache_dir: Optional[str] = None, token_set: Optional[TokenSet] = None, 
                 training_args: Optional[TrainingArguments] = None, model_args: Optional[ModelArguments] = None, 
                 text_normalizer: Callable[[str], str] = None, num_workers: Optional[int] = 1):
        """
        Finetune the model.

        Parameters
        ----------
        output_dir: str
            The output directory where the model checkpoints will be written.
        
        train_data: list[dict] = None
            A list of dict in the format {path: str, transcription: str} for training. 
            This parameter is optional only if data_cache_dir is specified and filled with already preprocessed data.
        
        eval_data: Optional[list[dict]] = None
            A list of dict in the format {path: str, transcription: str} for evaluation
        
        data_cache_dir: Optional[str] = None
            Pre-processed dataset cache directory. This can decrease the time needed to start the training (by using a lot of disk space).
        
        token_set: Optional[TokenSet] = None
            The token set to be used for training. This is mandatory if the model is not already fine-tuned.

        training_args: Optional[TrainingArguments] = None
            The training arguments.

        model_args: Optional[ModelArguments] = None
            The model arguments.

        text_normalizer: Callable[[str], str] = None
            Function used to normalize the transcriptions before evaluation.
        
        num_workers: Optional[int] = 1
            Number of workers to use for data loading.
        """

        if train_data is None and data_cache_dir is None:
            raise ValueError("train_data or data_cache_dir must be specified")

        if not self.is_finetuned and token_set is None:
            raise ValueError("The model is not fine-tuned yet, so you need to provide a token_set to fine-tune it")
        
        if self.is_finetuned:
            if token_set is not None:
                logger.warning("The model is already fine-tuned. So the provided token_set won't be used. The model's token_set will be used instead")
            token_set = self.token_set

        if text_normalizer is None:
            text_normalizer = DefaultTextNormalizer(token_set)

        if training_args is None:
            training_args = TrainingArguments()
        
        if model_args is None:
            model_args = ModelArguments()

        processor = self.processor if self.is_finetuned else token_set.to_processor(self.model_path)

        os.makedirs(output_dir, exist_ok=True)

        train_data_cache_dir = None
        eval_data_cache_dir = None

        if data_cache_dir is not None:
            train_data_cache_dir = os.path.join(data_cache_dir, "train")
            os.makedirs(train_data_cache_dir, exist_ok=True)

            eval_data_cache_dir = os.path.join(data_cache_dir, "eval")
            os.makedirs(eval_data_cache_dir, exist_ok=True)

        logger.info("Loading training data...")
        train_dataset = self._get_dataset(processor, text_normalizer, train_data, train_data_cache_dir, training_args.length_column_name, num_workers)
        
        eval_dataset = None
        if eval_data is not None or data_cache_dir is not None:
            logger.info("Loading evaluation data...")
            eval_dataset = self._get_dataset(processor, text_normalizer, eval_data, eval_data_cache_dir, training_args.length_column_name, num_workers)
        
        logger.info("Starting fine-tuning process...")
        
        finetune_ctc(self.model_path, output_dir, processor, train_dataset, eval_dataset, self.device, training_args, model_args)

        logger.info("Loading fine-tuned model...")

        self.model_path = output_dir
        self._load_model()



from __future__ import annotations
from typing import Optional, Union
import torch
import logging
from multiprocessing import Pool
import numpy as np
from huggingsound.token_set import TokenSet

class Decoder():
    """
    Decoder

    Parameters
    ----------
    token_set : TokenSet
        The TokenSet object to use for decoding.

    skip_special_tokens: Optional[bool] = True
        If True, skip the special tokens in the TokenSet during decoding.
    
    ms_per_timestep : Optional[int] = 20
        The number of milliseconds per timestep.

        The magic number 20 comes from the Wav2Vec2 convolutional layers. I'll try to explain this a little:
            The convolutional layers of the feature extractor have padding=0, kernel=(10,3,3,3,3,2,2) and strides=(5,2,2,2,2,2,2).
            So to map the model output (in timesteps) to the original audio (in milliseconds) we need to apply 
            the convolution output size formula ( ((input-kernel+2*padding)/stride)+1) ) on each layer, given some waveform input.
            However, I don't recommend that approach because it needs some extra computation and it's not really precise, 
            because there're a lot of overlappings due to the kernel and stride sizes desynchronization. 
            So a good and simple approximation can be extracted using just the convolutional layers strides, more specifically, 
            using the product of the strides:

            MS_PER_TIMESTEP = STRIDES_PRODUCT / WAVEFORM_POINTS_PER_MS
            MS_PER_TIMESTEP = 5*2^6 / 16 <- remember that the wav2vec input is in 16000Hz
            MS_PER_TIMESTEP = 20 <- the magic number :)

    probability_offset : Optional[float] = 1
        The probability offset to use when calculating the probability of a token when the end_timestep of a token isn't provided, 
        i. e., how many timesteps to look around to calc the probability.
    
    """

    def __init__(self, token_set: TokenSet, skip_special_tokens: Optional[bool] = True, ms_per_timestep: Optional[int] = 20, 
                 probability_offset: Optional[float] = 1):

        self.token_set = token_set
        self.skip_special_tokens = skip_special_tokens
        self.ms_per_timestep = ms_per_timestep
        self.probability_offset = probability_offset

    def _get_predictions(self, logits: torch.Tensor) -> list[dict]:
        """ 
        Get the predicted ids given the model's output logits.

        Parameters:
        ----------
            logits: torch.Tensor
                Model's output tensors of shape (BATCH_SIZE, TIMESTEPS, TOKEN_SET_SIZE)
        
        Returns:
        ----------
            list[dict]: prediction list of size BATCH_SIZE with format:
                [{
                    "ids": list,
                    "start_timesteps": list,
                    "end_timesteps": list,
                }, ...]
        """

        raise NotImplementedError()

    def _ctc_decode(self, predicted_ids: Union[torch.Tensor, np.ndarray, list[int]], return_timesteps: Optional[bool] = True) -> list[dict]:
        """ 
        Decode the predicted using the CTC decoding algorithm.

        Parameters:
        ----------
            predicted_ids: Union[torch.Tensor, np.ndarray, list[int]]
                Model's output tensors of shape (BATCH_SIZE, TIMESTEPS, TOKEN_SET_SIZE)
            
            return_timesteps: Optional[bool] = True
                If True, return the timesteps of the decoded sequence. 
        
        Returns:
        ----------
            list[dict]: decoded prediction list of size BATCH_SIZE with format:
                [{
                    "ids": list,
                    "start_timesteps": list,
                    "end_timesteps": list,
                }, ...]
        """

        predictions = []

        for i in range(len(predicted_ids)): # for each item in the batch
            
            i_predicted_ids = []
            i_start_timesteps = []
            i_end_timesteps = []
            previous_predicted_id = None
            
            for t in range(len(predicted_ids[i])): # for each timestep in the item

                predicted_id = int(predicted_ids[i][t])
                
                if predicted_id != self.token_set.blank_token_id:
                    
                    if len(i_predicted_ids) == 0 or previous_predicted_id == self.token_set.blank_token_id or predicted_id != i_predicted_ids[-1]:
                        i_predicted_ids.append(predicted_id)
                        if return_timesteps:
                            i_start_timesteps.append(t)
                            i_end_timesteps.append(t+1)

                    elif predicted_id == i_predicted_ids[-1] and return_timesteps:
                        i_end_timesteps[-1] = t

                previous_predicted_id = predicted_id
            
            predictions.append({
                "ids": i_predicted_ids,
                "start_timesteps": i_start_timesteps if return_timesteps else None,
                "end_timesteps": i_end_timesteps if return_timesteps else None,
            })

        return predictions

    def __call__(self, logits: torch.Tensor) -> list[dict]:
        """ 
        Getting the predictions given the model's output logits.

        Parameters:
        ----------
            logits: torch.Tensor
                Model's output tensors of shape (BATCH_SIZE, TIMESTEPS, TOKEN_SET_SIZE)
        
        Returns:
        ----------
            list[dict]: Decoded prediction list of size BATCH_SIZE with format:
                [{
                    "transcription": list,
                    "start_timesteps": list,
                    "end_timesteps": list,
                }, ...]
        """

        result = []
        
        predictions = self._get_predictions(logits)
        logits_probs = torch.nn.functional.softmax(logits.float(), dim=-1).to("cpu").detach()

        for i, prediction in enumerate(predictions):

            transcription = ""
            transcription_start_timestamps = []
            transcription_end_timestamps = []
            transcription_probabilities = []

            if "transcription" in prediction:
                transcription = prediction["transcription"]
                J = range(len(transcription))
            else:
                J = range(len(prediction["ids"]))

            for j in J:

                if "transcription" in prediction:
                    token = transcription[j] if transcription[j] != " " else self.token_set.silence_token
                    if token not in self.token_set.tokens:
                        token = self.token_set.unk_token
                    predicted_id = self.token_set.id_by_token[token]
                else:
                    predicted_id = prediction["ids"][j]

                    if predicted_id == self.token_set.silence_token_id:
                        transcription += " "
                    elif self.skip_special_tokens and self.token_set.tokens[predicted_id] in self.token_set.special_tokens:
                        continue
                    else:
                        transcription += self.token_set.tokens[predicted_id]
                
                if prediction["start_timesteps"] is not None:
                    start_timestep = prediction["start_timesteps"][j]
                    transcription_start_timestamps.append(int(start_timestep * self.ms_per_timestep))

                    # as we report the character based probability and more than one timestep can be responsable for the character prediction,
                    # when a start_timestep and end_timestep are provided we'll report the mean value of the this range of timesteps,
                    # otherwise we'll report the mean probability of a window defined be the start_timestep_t and start_timestep_t+1

                    if prediction["end_timesteps"] is not None:
                        window_end_timestep = prediction["end_timesteps"][j]
                    else:
                        window_end_timestep = start_timestep + 1
                    
                    if start_timestep == window_end_timestep: # it needs to have at least one timestep of difference
                        window_end_timestep += 1

                    window_probabilities = [x[predicted_id] for x in logits_probs[i][start_timestep:window_end_timestep]]
                    probability = float(np.mean(window_probabilities))
                    transcription_probabilities.append(probability)

                if prediction["end_timesteps"] is not None:
                    end_timestep = prediction["end_timesteps"][j]
                    transcription_end_timestamps.append(int(end_timestep * self.ms_per_timestep))

                    #probability = float(logits_probs[i][start_timestep][predicted_id])
                    #transcription_probabilities[-1] = probability

            # transcription trimming
            if len(transcription) > 0:
                
                left_offset = len(transcription) - len(transcription.lstrip())
                right_offset = len(transcription) - len(transcription.rstrip())

                transcription = transcription[left_offset:len(transcription)-right_offset]

                if len(transcription_start_timestamps) > 0:
                    transcription_start_timestamps = transcription_start_timestamps[left_offset:len(transcription_start_timestamps)-right_offset]
                if len(transcription_end_timestamps) > 0:
                    transcription_end_timestamps = transcription_end_timestamps[left_offset:len(transcription_end_timestamps)-right_offset]
                if len(transcription_probabilities) > 0:    
                    transcription_probabilities = transcription_probabilities[left_offset:len(transcription_probabilities)-right_offset]
            
            result.append({
                "transcription": transcription,
                "start_timestamps": transcription_start_timestamps if len(transcription_start_timestamps) > 0 else None,
                "end_timestamps": transcription_end_timestamps if len(transcription_end_timestamps) > 0 else None,
                "probabilities": transcription_probabilities if len(transcription_probabilities) > 0 else None,
            })
        
        return result


class GreedyDecoder(Decoder):
    """
    Greedy decoder

    Parameters
    ----------
    token_set : TokenSet
        The TokenSet object to use for decoding.
    """

    def __init__(self, token_set: TokenSet):
        super().__init__(token_set)

    def _get_predictions(self, logits: torch.Tensor):

        predicted_ids = torch.argmax(logits, dim=-1)
        predictions = self._ctc_decode(predicted_ids)

        return predictions


class ParlanceLMDecoder(Decoder):
    """
    Parlance Language Model decoder

    Parameters
    ----------
    token_set : TokenSet
        The TokenSet object to use for decoding.
    
    lm_path : str
        Path to the KenLM language model file
    
    alpha: Optional[float] = 2.0
        Weighting associated with the LMs probabilities. A weight of 0 means the LM has no effect.
    
    beta: Optional[float] = -1.0
        Weight associated with the number of words within our beam (LM usage reward).
    
    cutoff_top_n: Optional[int] = 40
        Cutoff number in pruning. Only the top cutoff_top_n characters with the highest probability in 
        the TokenSet will be used in beam search.
    
    cutoff_prob: Optional[float] = 1.0
        Cutoff probability in pruning. 1.0 means no pruning.
    
    beam_width: Optional[int] = 100
        This controls how broad the beam search is. Higher values are more likely to find top beams, 
        but they also will make your beam search exponentially slower. Furthermore, the longer your outputs, 
        the more time large beams will take. This is an important parameter that represents a tradeoff you need 
        to make based on your dataset and needs.
    
    num_processes: Optional[int] = 4
        Parallelize the batch using num_processes workers. You probably want to pass the number of cpus your computer has. 
        You can find this in python with import multiprocessing then n_cpus = multiprocessing.cpu_count()
    """

    def __init__(self, token_set: TokenSet, lm_path: str, alpha: Optional[float] = 2.0, beta: Optional[float] = -1.0, 
                 cutoff_top_n: Optional[int] = 40, cutoff_prob: Optional[float] = 1.0, beam_width: Optional[int] = 100, 
                 num_processes: Optional[int] = 4):
        
        super().__init__(token_set)

        self.lm_path = lm_path
        self.alpha = alpha
        self.beta = beta
        self.cutoff_top_n = cutoff_top_n
        self.cutoff_prob = cutoff_prob
        self.beam_width = beam_width
        self.num_processes = num_processes

        try:
            from ctcdecode import CTCBeamDecoder
        except ImportError:
            raise ImportError("To use this decoder please install the ctcdecoder from https://github.com/parlance/ctcdecode")

        # creating the tokens forcing the silence token to be a whitespace
        tokens = [x if x != self.token_set.silence_token else " " for x in self.token_set.tokens]
        
        self.ctcdecoder = CTCBeamDecoder(tokens, 
            model_path=self.lm_path,
            alpha=self.alpha,
            beta=self.beta,
            cutoff_top_n=self.cutoff_top_n,
            cutoff_prob=self.cutoff_prob,
            beam_width=self.beam_width,
            num_processes=self.num_processes,
            blank_id=self.token_set.blank_token_id,
            
            # If your outputs have passed through a softmax and represent probabilities, this should be false, 
            # if they passed through a LogSoftmax and represent negative log likelihood, you need to pass True. 
            # If you don't understand this, run print(output[0][0].sum()), if it's a negative number you've probably got NLL 
            # and need to pass True, if it sums to ~1.0 you should pass False. We set this to True by default, 
            # 'cause nowadays the logits are not normalized and we'll apply a softmax to them before decoding.
            log_probs_input=True
        )

    def _get_predictions(self, logits: torch.Tensor) -> list[dict]:

        # beam_results (BATCHSIZE x N_BEAMS X N_TIMESTEPS): A batch containing the series 
        # of characters (these are ints, you still need to decode them back to your text) representing 
        # results from a given beam search. Note that the beams are almost always shorter than the 
        # total number of timesteps, and the additional data is non-sensical, so to see the top beam 
        # (as int labels) from the first item in the batch, you need to run beam_results[0][0][:out_len[0][0]].

        # beam_scores (BATCHSIZE x N_BEAMS): A batch with the approximate CTC score of each beam 
        # If this is true, you can get the model's confidence that the beam is correct with p=1/np.exp(beam_score).

        # beam_timesteps (BATCHSIZE x N_BEAMS X N_TIMESTEPS) the timestep at which the nth output character has peak probability. 
        # Can be used as alignment between the audio and the transcript.

        # beam_end_lens (BATCHSIZE x N_BEAMS): out_lens[i][j] is the length of the jth beam_result, of item i of your batch.
        
        beam_results, beam_scores, beam_timesteps, beam_end_lens = self.ctcdecoder.decode(logits.to("cpu"))

        predictions = []

        B, N, T = beam_results.size()

        for i in range(B):
            length = beam_end_lens[i][0]

            predictions.append({
                "ids": beam_results[i][0][:length],
                "start_timesteps": beam_timesteps[i][0][:length],
                "end_timesteps": None,
            })

        return predictions


class FlashlightLMDecoder(Decoder):
    """
    Flashlight Language Model decoder

    Parameters
    ----------
    token_set : TokenSet
        The TokenSet object to use for decoding.
    
    lm_path : str
        Path to the KenLM language model file
    
    lexicon_path: Optional[str] = None
        Path to the lexicon file
    
    beam_size: Optional[int] = 100
        Number of top hypothesis to preserve at each decoding step.
    
    beam_threshold: Optional[float] = 100.0
        Preserve a hypothesis only if its score is not far away from the current best hypothesis score.
    
    beam_size_token: Optional[int] = None
        Restrict number of tokens by top am scores (if you have a huge token set). If None, the beam size is len(TokenSet)
    
    lm_weight: Optional[float] = 2.0
        Language model weight for LM score.
    
    word_score: Optional[float] = -1.0
        Score for words appearance in the transcription (word insertion penalty).
    
    sil_score: Optional[float] = 0.0
        Score for silence appearance in the transcription.
    
    decoder_device: Optional[str] = "cpu"
        The device where the decoder will be executed, by default it's "cpu", if you want to use gpu, 
        then you'll need to install the flashlight with CUDA support and set the decoder_device to "cuda"
    
    """

    def __init__(self, token_set: TokenSet, lm_path: str, lexicon_path: Optional[str] = None, beam_size: Optional[int] = 100, 
                 beam_threshold: Optional[float] = 100.0, beam_size_token: Optional[int] = None, lm_weight: Optional[float] = 2.0,
                 word_score: Optional[float] = -1.0, sil_score: Optional[float] = 0.0, decoder_device: Optional[str] = "cpu"):
        
        super().__init__(token_set)
                
        try:
            from flashlight.lib.text.dictionary import create_word_dict, load_words
            from flashlight.lib.text.decoder import (
                CriterionType,
                LexiconDecoderOptions,
                KenLM,
                SmearingMode,
                Trie,
                LexiconDecoder,
            )
            from flashlight.lib.text.decoder import LexiconFreeDecoder, LexiconFreeDecoderOptions

        except:
            raise ImportError("To use this decoder please install the flashlight python bindings from https://github.com/facebookresearch/flashlight/tree/master/bindings/python")
        
        self.lm_path = lm_path
        self.lexicon_path = lexicon_path
        self.beam_size = beam_size
        self.beam_threshold = beam_threshold
        self.beam_size_token = beam_size_token if beam_size_token is not None else self.token_set.size
        self.lm_weight = lm_weight
        self.word_score = word_score
        self.sil_score = sil_score
        self.decoder_device = decoder_device

        if self.lexicon_path is not None:
            tokens_keys = self.token_set.id_by_token.keys()

            # load lexicon file, which defines spelling of words
            # the format word and its tokens spelling separated by the spaces, e.g.:
            # ann a n 1 |
            self.lexicon = load_words(self.lexicon_path)
            
            # read lexicon and store it in the w2l dictionary
            self.word_dict = create_word_dict(self.lexicon)

            # create Kenlm language model
            self.lm = KenLM(self.lm_path, self.word_dict)
            
            # build trie (specifying how many tokens we have and silence index)
            # Trie is necessary to do beam-search decoding with word-level lm
            # We restrict our search only to the words from the lexicon
            # Trie is constructed from the lexicon, each node is a token
            # path from the root to a leaf corresponds to a word spelling in the lexicon
            self.trie = Trie(self.token_set.size, self.token_set.silence_token_id)
            start_state = self.lm.start(False)

            # use heuristic for the trie, called smearing:
            # predict lm score for each word in the lexicon, set this score to a leaf
            # (we predict lm score for each word as each word starts a sentence)
            # word score of a leaf is propagated up to the root to have some proxy score
            # for any intermediate path in the trie
            # SmearingMode defines the function how to process scores
            # in a node came from the children nodes:
            # could be max operation or logadd or none
            for i, (word, spellings) in enumerate(self.lexicon.items()):
                word_idx = self.word_dict.get_index(word)
                _, score = self.lm.score(start_state, word_idx)

                for spelling in spellings:
                    spelling_idxs = []
                    for token in spelling:
                        if token.upper() in tokens_keys:
                            spelling_idxs.append(self.token_set.id_by_token[token.upper()])
                        elif token.lower() in tokens_keys:
                            spelling_idxs.append(self.token_set.id_by_token[token.lower()])
                        else:
                            logging.warning(f"WARNING: The token {token} not exist in your TokenSet, using <unk> token instead")
                            spelling_idxs.append(self.token_set.unk_token_id)
                        
                    self.trie.insert(spelling_idxs, word_idx, score)
            self.trie.smear(SmearingMode.MAX)

            # Define decoder options
            self.decoder_opts = LexiconDecoderOptions(
                beam_size=self.beam_size,
                beam_size_token=self.beam_size_token,
                beam_threshold=self.beam_threshold,
                lm_weight=self.lm_weight,
                word_score=self.word_score,
                unk_score=-np.inf,
                sil_score=self.sil_score,
                log_add=False,
                criterion_type=CriterionType.CTC,
            )

            # define lexicon beam-search decoder with word-level lm
            # LexiconDecoder(decoder options, trie, lm, silence index,
            #                blank index (for CTC), unk index,
            #                transitiona matrix, is token-level lm)
            self.decoder = LexiconDecoder(
                self.decoder_opts,
                self.trie,
                self.lm,
                self.token_set.silence_token_id,
                self.token_set.blank_token_id,
                self.token_set.unk_token_id,
                [],
                False,
            )

        else:
            
            # NOTE: lexicon free decoding can only be properly done with a unit language model
            
            # The general format for each line of the lexicon file is:
            # {representation of the word/unit in the language model}      {representation of ctc model labels used for training}
            
            # So if your model was trained on letters, then the format is:
            # WRITE        W R I T E |
            
            # But if it is phones, it could be:
            # WRITE        WRY  T |

            # And if it's just words, it's basically the identity function:
            # WRITE      WRITE

            # ... so when the representation of the word/unit is the same as the representation of the ctc labels,
            # you have a unit language model, and you can use a lexicon free decoding, 
            # but this decoder usually doesn't give good results, so I don't recommend to use it.

            d = {w: [[w]] for w in self.token_set.id_by_token.keys()}
            self.word_dict = create_word_dict(d)
            self.lm = KenLM(self.lm_path, self.word_dict)
            self.decoder_opts = LexiconFreeDecoderOptions(
                beam_size=self.beam_size,
                beam_size_token=self.beam_size_token,
                beam_threshold=self.beam_threshold,
                lm_weight=self.lm_weight,
                sil_score=self.sil_score,
                log_add=False,
                criterion_type=CriterionType.CTC,
            )
            self.decoder = LexiconFreeDecoder(
                self.decoder_opts, 
                self.lm, 
                self.token_set.silence_token_id, 
                self.token_set.blank_token_id, 
                []
            )

    def _get_predictions(self, logits: torch.Tensor) -> list[dict]:

        # This parameter controls the number of best results to be considered for decoding. 
        # Nowadays, it's 1 by default. In the future, I may change this if I add some randomness for decoding.
        nbest_results = 1
        
        logits = logits.to(self.decoder_device)
        B, T, N = logits.size()

        tokens = []
        # scores = []
        for i in range(B):
            logits_ptr = logits.data_ptr() + 4 * i * logits.stride(0)

            # decoder.decode(emissions, Time, Ntokens)
            # result is a list of sorted hypothesis, 0-index is the best hypothesis
            # each hypothesis is a struct with "score" and "words" representation
            # in the hypothesis and the "tokens" representation
            results = self.decoder.decode(logits_ptr, T, N)

            best_results = results[: nbest_results] # getting the best candidates
            tokens_best = []
            # scores_best = []
            for result in best_results:
                tokens_best.append(result.tokens)
                # scores_best.append(result.score)
            tokens.append(tokens_best)
            # scores.append(scores_best)

        token_array = np.array(tokens, dtype=object).transpose((1, 0, 2))
        # scores_arrray = np.array(scores, dtype=object).transpose()

        predicted_ids = token_array[0][:] # getting best candidates

        predictions = self._ctc_decode(predicted_ids)

        return predictions


class KenshoLMDecoder(Decoder):
    """
    Kensho-technologies' pyctcdecode Language Model decoder

    decoder = build_ctcdecoder(
        labels,
        kenlm_model,
        alpha=0.5,  # tuned on a val set
        beta=1.0,  # tuned on a val set
    )
    text = decoder.decode(logits)

    labels: class containing the labels for input logit matrices
    kenlm_model_path: path to kenlm n-gram language model
    unigrams: list of known word unigrams
    alpha: weight for language model during shallow fusion
    beta: weight for length score adjustment of during scoring
    unk_score_offset: amount of log score offset for unknown tokens
    lm_score_boundary: whether to have kenlm respect boundaries when scoring

    Parameters
    ----------
    token_set : TokenSet
        The TokenSet object to use for decoding.
    
    lm_path : str
        Path to the KenLM language model file
    
    alpha: Optional[float] = 2.0
        Weighting associated with the LMs probabilities. A weight of 0 means the LM has no effect.
    
    beta: Optional[float] = -1.0
        Weight associated with the number of words within our beam (LM usage reward).

    unigrams_path: Optional[str] = None
        The path of the unigrams file.

    unk_score_offset: Optional[float] = -10.0
        Amount of log score offset for unknown tokens.

    lm_score_boundary: Optional[bool] = True
        Whether to have kenlm respect boundaries when scoring.

    beam_width: Optional[int] = 100
        Maximum number of beams at each step in decoding.

    beam_prune_logp: Optional[float] = -10.0
        Beams that are much worse than best beam will be pruned.

    token_min_logp: Optional[float] = -5.0
        Tokens below this logp are skipped unless they are argmax of frame.

    prune_history: Optional[bool] = False
        Whether to filter out beams that are the same over max_ngram history.

    hotwords: Optional[list[str]] = None
        List of words with extra importance, can be OOV for LM.

    hotword_weights: Optional[float]) = 10.0
        Weight factor for hotword importance.

    """

    def __init__(self, token_set: TokenSet, lm_path: str, alpha: Optional[float] = 2.0, beta: Optional[float] = -1.0, 
                 unigrams_path: Optional[str] = None, unk_score_offset: Optional[float] = -10.0, lm_score_boundary: Optional[bool] = True,
                 beam_width: Optional[int] = 100, beam_prune_logp: Optional[float] = -10.0, token_min_logp: Optional[float] = -5.0,
                 prune_history: Optional[bool] = False, hotwords: Optional[list[str]] = None, hotword_weights: Optional[float] = 10.0):
        
        super().__init__(token_set)

        self.lm_path = lm_path
        self.alpha = alpha
        self.beta = beta
        self.unigrams_path = unigrams_path
        self.unk_score_offset = unk_score_offset
        self.lm_score_boundary = lm_score_boundary
        self.beam_width = beam_width
        self.beam_prune_logp = beam_prune_logp
        self.token_min_logp = token_min_logp
        self.prune_history = prune_history
        self.hotwords = hotwords
        self.hotword_weights = hotword_weights

        try:
            from pyctcdecode import build_ctcdecoder
        except ImportError:
            raise ImportError("To use this decoder please install the pyctcdecode from https://github.com/kensho-technologies/pyctcdecode")

        if self.unigrams_path is not None:
            self.unigrams = []
            with open(self.unigrams_path, "r") as f:
                for line in f:
                    self.unigrams.append(line.strip())

        # creating the tokens forcing the silence token to be a whitespace
        tokens = [x if x != self.token_set.silence_token else " " for x in self.token_set.tokens]
        
        self.decoder = build_ctcdecoder(
            tokens,
            self.lm_path,
            alpha=self.alpha,
            beta=self.beta,
            unigrams=self.unigrams,
            unk_score_offset=self.unk_score_offset,
            lm_score_boundary=self.lm_score_boundary
        )

    def _get_predictions(self, logits: torch.Tensor) -> list[dict]:

        with Pool() as pool:

            decoder_outputs = self.decoder.decode_beams_batch(
                pool, logits.cpu().detach().numpy(), beam_width=self.beam_width, beam_prune_logp=self.beam_prune_logp, 
                token_min_logp=self.token_min_logp, prune_history=self.prune_history, 
                hotwords=self.hotwords, hotword_weight=self.hotword_weights
            )

            predictions = []

            for decoder_output in decoder_outputs:

                transcription, word_frames, logit_score, lm_score = decoder_output[0]
                start_timesteps = []
                last_end_timestep = None
                
                for word_frame in word_frames:
                    word = word_frame[0]
                    start_timestep = word_frame[1][0]
                    end_timestep = word_frame[1][1]

                    # as the pyctcdecode doesn't return the character based timestamp, we need to make an approximation of it
                    timestep_per_char = max(1, int((end_timestep - start_timestep)/len(word)))

                    # handling whitespaces
                    if last_end_timestep is not None:
                        start_timesteps.append(last_end_timestep)

                    for i in range(0, len(word)):
                        start_timesteps.append(start_timestep + i*timestep_per_char)

                    last_end_timestep = end_timestep

                predictions.append({
                    "transcription": transcription,
                    "start_timesteps": start_timesteps,
                    "end_timesteps": None,
                })

        return predictions



from __future__ import annotations
import warnings
import librosa
import numpy as np
from typing import Iterator, Optional
from datasets import Dataset
from collections import Counter


def get_chunks(values: list, n: int) -> Iterator:
    """ 
    Yield successive n-sized chunks from values.

    Parameters:
    ----------
        values: list
            values to be chunked
       
        n: int
            chunk size

    Returns:
    ----------
        Iterator: A chunk iterator
    """

    for i in range(0, len(values), n):
        yield values[i:i + n]


def get_waveforms(pahts: list[str], sampling_rate: Optional[int] = 16000) -> list[np.ndarray]:
    """ 
    Get waveforms from audio files.

    Parameters:
    ----------
        pahts: list[str]
            paths to audio files
        
        sampling_rate: Optional[int] = 16000
            sampling rate of waveforms

    Returns:
    ----------
        list[np.ndarray]: waveforms from audio files
    """

    waveforms = []
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        for path in pahts:
            waveform, sr = librosa.load(path, sr=sampling_rate)
            waveforms.append(waveform)
    
    return waveforms


def get_dataset_from_dict_list(data: list[dict]) -> Dataset:
    """ 
    Get dataset from dict list.

    Parameters:
    ----------
        data: list[dict]
            dict list to be converted to dataset

    Returns:
    ----------
        Dataset: dataset from dict list
    """

    keys = data[0].keys()
    transformed_data = {}
    for key in keys:
        for d in data:
            transformed_data[key] = transformed_data.get(key, []) + [d[key]]

    dataset = Dataset.from_dict(transformed_data)

    return dataset


class CustomCounter(Counter):
    """ The operation 'Counter() + Counter()' ignore keys with zero values, this Counter extension fix that"""
    
    def __add__(self, other):
        if not isinstance(other, Counter):
            return NotImplemented
        result = CustomCounter()
        for elem, count in self.items():
            newcount = count + other[elem]
            result[elem] = newcount
        for elem, count in other.items():
            if elem not in self:
                result[elem] = count
        return result


from __future__ import annotations
from typing import Optional
import jiwer
import gc

def wer(predictions: list[str], references: list[str], chunk_size: Optional[int] = None) -> float:
    """
    Computes WER score of transcribed segments against references.

    Word error rate (WER) is a common metric of the performance of an automatic speech recognition system.
    The general difficulty of measuring performance lies in the fact that the recognized word sequence can 
    have a different length from the reference word sequence (supposedly the correct one). 
    The WER is derived from the Levenshtein distance, working at the word level instead of the phoneme level.

    Word error rate can then be computed as:
    WER = (S + D + I) / N = (S + D + I) / (S + D + C)
    where
    S is the number of substitutions,
    D is the number of deletions,
    I is the number of insertions,
    C is the number of correct words,
    N is the number of words in the reference (N=S+D+C).

    WER's output is always a number between 0 and 1. This value indicates the percentage of words that were incorrectly predicted. 
    The lower the value, the better the performance of the ASR system with a WER of 0 being a perfect score.
    
    Parameters:
    ----------
        predictions: list[str]
            List of transcribtions to score.
        
        references: list[str]
            List of references for each speech input.
        
        chunk_size: Optional[int] = None
            Size of the chunk to use for computation. 
            When this value is specified, the function will chunk the data into batches of the specified size and compute the WER on each batch.
            After all batches are computed, the function will compute the average WER over all batches.
            (You will probably need to define this if you have memory issues).
    
    Returns:
    ----------
        float: the word error rate
    """
    
    if chunk_size is None: return jiwer.wer(references, predictions)
    
    start = 0
    end = chunk_size
    H, S, D, I = 0, 0, 0, 0

    while start < len(references):

        chunk_metrics = jiwer.compute_measures(references[start:end], predictions[start:end], 
                                               truth_transform=jiwer.transformations.wer_default, 
                                               hypothesis_transform=jiwer.transformations.wer_default)
        H = H + chunk_metrics["hits"]
        S = S + chunk_metrics["substitutions"]
        D = D + chunk_metrics["deletions"]
        I = I + chunk_metrics["insertions"]
        start += chunk_size
        end += chunk_size

        # sometimes this metric uses a lot of memory, so we'll try to free it here
        del chunk_metrics
        gc.collect()
    
    return float(S + D + I) / float(H + S + D)


def cer(predictions: list[str], references: list[str], chunk_size: Optional[int] = None) -> float:
    """
    Computes CER score of transcribed segments against references.

    Character error rate (CER) is a common metric of the performance of an automatic speech recognition system.
    CER is similar to Word Error Rate (WER), but operate on character insted of word. Please refer to docs of WER for further information.
    
    Character error rate can be computed as:
    CER = (S + D + I) / N = (S + D + I) / (S + D + C)
    where
    S is the number of substitutions,
    D is the number of deletions,
    I is the number of insertions,
    C is the number of correct characters,
    N is the number of characters in the reference (N=S+D+C).

    CER's output is always a number between 0 and 1. This value indicates the percentage of characters that were incorrectly predicted. 
    The lower the value, the better the performance of the ASR system with a CER of 0 being a perfect score.
    
    Parameters:
    ----------
        predictions: list[str]
            List of transcribtions to score.
        
        references: list[str]
            List of references for each speech input.
        
        chunk_size: Optional[int] = None
            Size of the chunk to use for computation. 
            When this value is specified, the function will chunk the data into batches of the specified size and compute the CER on each batch.
            After all batches are computed, the function will compute the average CER over all batches.
            (You will probably need to define this if you have memory issues).
   
    Returns:
    ----------
        float: the character error rate
    """

    if chunk_size is None: return jiwer.cer(references, predictions)

    start = 0
    end = chunk_size
    H, S, D, I = 0, 0, 0, 0

    while start < len(references):
        chunk_metrics = jiwer.cer(references[start:end], predictions[start:end], return_dict=True)
        H = H + chunk_metrics["hits"]
        S = S + chunk_metrics["substitutions"]
        D = D + chunk_metrics["deletions"]
        I = I + chunk_metrics["insertions"]
        start += chunk_size
        end += chunk_size

        # sometimes this metric uses a lot of memory, so we'll try to free it here
        del chunk_metrics
        gc.collect()
    
    return float(S + D + I) / float(H + S + D)



from __future__ import annotations
import tempfile
import os
import json
import logging
from typing import Optional
from transformers import Wav2Vec2Processor, AutoConfig, AutoTokenizer, AutoFeatureExtractor

class TokenSet():
    """
    TokenSet

    Parameters
    ----------
    tokens : list[str]
        List of tokens.
    
    blank_token : Optional[str] = "<pad>"
        Blank token
    
    silence_token : Optional[str] = "|"
        Silence token
    
    unk_token : Optional[str] = "<unk>"
        Unk token
    
    bos_token : Optional[str] = "<s>"
        BOS token

    eos_token : Optional[str] = "</s>"
        EOS token

    letter_case: str
        Case mode to be applied to the transcription, can be 'lowercase', 'uppercase' or None (None == keep the original letter case). Default is None.

    """
    
    def __init__(self, tokens: list[str], blank_token: Optional[str] = "<pad>", silence_token: Optional[str] = "|", unk_token: Optional[str] = "<unk>", 
                 bos_token: Optional[str] = "<s>", eos_token: Optional[str] = "</s>", letter_case: str = "lowercase"):

        self.tokens = tokens
        self.blank_token = blank_token
        self.silence_token = silence_token
        self.unk_token = unk_token
        self.bos_token = bos_token
        self.eos_token = eos_token
        self.letter_case = letter_case

        if self.letter_case == "lowercase":
            self.tokens = [token.lower() if token not in self.special_tokens else token for token in self.tokens]
        elif self.letter_case == "uppercase":
            self.tokens = [token.upper() if token not in self.special_tokens else token for token in self.tokens]
        
        if blank_token not in tokens:
            logging.warning(f"blank_token {blank_token} not in provided tokens. It will be added to the list of tokens")
            self.tokens.append(blank_token)

        if silence_token not in tokens:
            logging.warning(f"silence_token {silence_token} not in provided tokens. It will be added to the list of tokens")
            self.tokens.append(silence_token)
        
        if unk_token not in tokens:
            logging.warning(f"unk_token {unk_token} not in provided tokens. It will be added to the list of tokens")
            self.tokens.append(unk_token)

        if bos_token not in tokens:
            logging.warning(f"bos_token {bos_token} not in provided tokens. It will be added to the list of tokens")
            self.tokens.append(bos_token)
        
        if eos_token not in tokens:
            logging.warning(f"eos_token {eos_token} not in provided tokens. It will be added to the list of tokens")
            self.tokens.append(eos_token)

        self.id_by_token = {token: i for i, token in enumerate(self.tokens)}

    @property
    def blank_token_id(self):
        return self.id_by_token[self.blank_token]

    @property
    def silence_token_id(self):
        return self.id_by_token[self.silence_token]
    
    @property
    def unk_token_id(self):
        return self.id_by_token[self.unk_token]

    @property
    def bos_token_id(self):
        return self.id_by_token[self.bos_token]

    @property
    def eos_token_id(self):
        return self.id_by_token[self.eos_token]

    @property
    def non_special_tokens(self):
        return [token for token in self.tokens if token not in self.special_tokens]

    @property
    def special_tokens(self):
        return [self.blank_token, self.silence_token, self.unk_token, self.bos_token, self.eos_token]

    @property
    def size(self):
        return len(self.tokens)

    def to_processor(self, model_name_or_path: str="facebook/wav2vec2-large-xlsr-53"):

        tokens_dict = {v: i for i, v in enumerate(self.tokens)}

        with tempfile.TemporaryDirectory() as tmpdirname:

            vocab_path = os.path.join(tmpdirname, "vocab.json")

            with open(vocab_path, "w") as vocab_file:
                json.dump(tokens_dict, vocab_file)

            config = AutoConfig.from_pretrained(model_name_or_path)
            config_for_tokenizer = config if config.tokenizer_class is not None else None
            tokenizer_type = config.model_type if config.tokenizer_class is None else None

            tokenizer = AutoTokenizer.from_pretrained(
                tmpdirname,
                config=config_for_tokenizer,
                tokenizer_type=tokenizer_type,
                bos_token=self.bos_token,
                eos_token=self.eos_token,
                unk_token=self.unk_token,
                pad_token=self.blank_token,
                word_delimiter_token=self.silence_token,
                do_lower_case=False,
                # do_lower_case=self.letter_case == "lowercase",
                #TODO: fix transformers/models/wav2vec2/tokenization_wav2vec2.py:199
            )

            feature_extractor = AutoFeatureExtractor.from_pretrained(model_name_or_path)

            return Wav2Vec2Processor(feature_extractor=feature_extractor, tokenizer=tokenizer)

    @classmethod
    def from_processor(cls, processor: Wav2Vec2Processor, letter_case: str = "lowercase"):

        blank_token = processor.tokenizer.pad_token
        silence_token = processor.tokenizer.word_delimiter_token
        unk_token = processor.tokenizer.unk_token
        bos_token = processor.tokenizer.bos_token
        eos_token = processor.tokenizer.eos_token
        tokens = [x for x in processor.tokenizer.convert_ids_to_tokens(range(0, processor.tokenizer.vocab_size))]

        return cls(tokens, blank_token, silence_token, unk_token, bos_token, eos_token, letter_case)

    def save(self, path: str):

        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.__dict__, f, indent=2, ensure_ascii=False)

    @classmethod
    def load(cls, path: str):

        with open(path, encoding="utf-8") as f:
            o = json.load(f)
            return cls(o["tokens"], o["blank_token"], o["silence_token"], o["unk_token"], o["bos_token"], o["eos_token"], o["letter_case"])


from __future__ import annotations
import re
from huggingsound.token_set import TokenSet


class DefaultTextNormalizer():
    """
    Default Text Normalizer.

    Parameters
    ----------
    token_set: TokenSet
        TokenSet.
    """

    def __init__(self, token_set: TokenSet):
        
        self.token_set = token_set
        self.valid_tokens = self.token_set.non_special_tokens

        if self.token_set.letter_case == "lowercase":
            self.valid_tokens = set([x.lower() for x in self.valid_tokens])
        elif self.token_set.letter_case == "uppercase":
            self.valid_tokens = set([x.upper() for x in self.valid_tokens])

        self.invalid_chars_regex = f"[^\s{re.escape(''.join(set(self.valid_tokens)))}]"

    def __call__(self, sentence: str) -> str:
        """ 
        Preprocess the sentence for training/evaluation.

        Parameters:
        ----------
            sentence: str
                The sentence to be preprocessed.

        Returns:
        ----------
            str: The preprocessed sentence.
        """
        
        # to (lower|upper)case case
        if self.token_set.letter_case == "lowercase":
            sentence = sentence.lower()
        elif self.token_set.letter_case == "uppercase":
            sentence = sentence.upper()

        # removing invalid characters
        sentence = re.sub(self.invalid_chars_regex, " ", sentence)
        sentence = re.sub("\s+", " ", sentence).strip()
        
        return sentence



from huggingsound.trainer import TrainingArguments, ModelArguments
from huggingsound.speech_recognition.model import SpeechRecognitionModel
from huggingsound.speech_recognition.decoder import Decoder, GreedyDecoder, ParlanceLMDecoder, FlashlightLMDecoder, KenshoLMDecoder
from huggingsound.token_set import TokenSet
from huggingsound.normalizer import DefaultTextNormalizer



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['huggingsound',
 'huggingsound.speech_classification',
 'huggingsound.speech_recognition']

package_data = \
{'': ['*']}

install_requires = \
['datasets>=1.18.3,<2.0.0',
 'jiwer>=2.3.0,<3.0.0',
 'librosa>=0.8.1,<0.9.0',
 'llvmlite>=0.36.0,<0.37.0',
 'numba>=0.53.1,<0.54.0',
 'torch>=1.9.0,<2.0.0',
 'transformers>=4.16.2,<5.0.0']

setup_kwargs = {
    'name': 'huggingsound',
    'version': '0.1.1',
    'description': "HuggingSound: A toolkit for speech-related tasks based on HuggingFace's tools.",
    'long_description': '# HuggingSound\n\nHuggingSound: A toolkit for speech-related tasks based on [HuggingFace\'s](https://huggingface.co/) tools.\n\nI have no intention of building a very complex tool here. \nI just wanna have an easy-to-use toolkit for my speech-related experiments.\nI hope this library could be helpful for someone else too :)\n\n# Requirements\n\n- Python 3.7+\n\n# Installation\n\n```console\n$ pip install huggingsound\n```\n\n# How to use it?\n\nI\'ll try to summarize the usage of this toolkit. \nBut many things will be missing from the documentation below. I promise to make it better soon.\nFor now, you can open an issue if you have some questions or look at the source code to see how it works.\nYou can check more usage examples in the repository `examples` folder.\n\n## Speech recognition\n\nFor speech recognition you can use any CTC model hosted on the Hugging Face Hub. You can find some available models [here](https://huggingface.co/models?pipeline_tag=automatic-speech-recognition).\n\n### Inference\n\n```python\nfrom huggingsound import SpeechRecognitionModel\n\nmodel = SpeechRecognitionModel("jonatasgrosman/wav2vec2-large-xlsr-53-english")\naudio_paths = ["/path/to/sagan.mp3", "/path/to/asimov.wav"]\n\ntranscriptions = model.transcribe(audio_paths)\n\nprint(transcriptions)\n\n# transcriptions format (a list of dicts, one for each audio file):\n# [\n#  {\n#   "transcription": "extraordinary claims require extraordinary evidence", \n#   "start_timestamps": [100, 120, 140, 180, ...],\n#   "end_timestamps": [120, 140, 180, 200, ...],\n#   "probabilities": [0.95, 0.88, 0.9, 0.97, ...]\n# },\n# ...]\n#\n# as you can see, not only the transcription is returned but also the timestamps (in milliseconds) \n# and probabilities of each character of the transcription.\n\n```\n\n### Inference (boosted by a language model)\n\n```python\nfrom huggingsound import SpeechRecognitionModel, KenshoLMDecoder\n\nmodel = SpeechRecognitionModel("jonatasgrosman/wav2vec2-large-xlsr-53-english")\naudio_paths = ["/path/to/sagan.mp3", "/path/to/asimov.wav"]\n\n# The LM format used by the LM decoders is the KenLM format (arpa or binary file).\n# You can download some LM files examples from here: https://huggingface.co/jonatasgrosman/wav2vec2-large-xlsr-53-english/tree/main/language_model\nlm_path = "path/to/your/lm_files/lm.binary"\nunigrams_path = "path/to/your/lm_files/unigrams.txt"\n\n# We implemented three different decoders for LM boosted decoding: KenshoLMDecoder, ParlanceLMDecoder, and FlashlightLMDecoder\n# On this example, we\'ll use the KenshoLMDecoder\n# To use this decoder you\'ll need to install the Kensho\'s ctcdecode first (https://github.com/kensho-technologies/pyctcdecode)\ndecoder = KenshoLMDecoder(model.token_set, lm_path=lm_path, unigrams_path=unigrams_path)\n\ntranscriptions = model.transcribe(audio_paths, decoder=decoder)\n\nprint(transcriptions)\n\n```\n\n### Evaluation\n```python\nfrom huggingsound import SpeechRecognitionModel\n\nmodel = SpeechRecognitionModel("jonatasgrosman/wav2vec2-large-xlsr-53-english")\n\nreferences = [\n    {"path": "/path/to/sagan.mp3", "transcription": "extraordinary claims require extraordinary evidence"},\n    {"path": "/path/to/asimov.wav", "transcription": "violence is the last refuge of the incompetent"},\n]\n\nevaluation = model.evaluate(references)\n\nprint(evaluation)\n\n# evaluation format: {"wer": 0.08, "cer": 0.02}\n```\n\n### Fine-tuning\n```python\nfrom huggingsound import TrainingArguments, ModelArguments, SpeechRecognitionModel, TokenSet\n\nmodel = SpeechRecognitionModel("facebook/wav2vec2-large-xlsr-53")\noutput_dir = "my/finetuned/model/output/dir"\n\n# first of all, you need to define your model\'s token set\ntokens = ["a", "b", "c", "d", "e", "f", "g", "h", "i", "j", "k", "l", "m", "n", "o", "p", "q", "r", "s", "t", "u", "v", "w", "x", "y", "z", "\'"]\ntoken_set = TokenSet(tokens)\n\n# define your train/eval data\ntrain_data = [\n    {"path": "/path/to/sagan.mp3", "transcription": "extraordinary claims require extraordinary evidence"},\n    {"path": "/path/to/asimov.wav", "transcription": "violence is the last refuge of the incompetent"},\n]\neval_data = [\n    {"path": "/path/to/sagan2.mp3", "transcription": "absence of evidence is not evidence of absence"},\n    {"path": "/path/to/asimov2.wav", "transcription": "the true delight is in the finding out rather than in the knowing"},\n]\n\n# and finally, fine-tune your model\nmodel.finetune(\n    output_dir, \n    train_data=train_data, \n    eval_data=eval_data, # the eval_data is optional\n    token_set=token_set, \n)\n\n```\n\n# Troubleshooting\n\n- If you are having trouble when loading MP3 files: `$ sudo apt-get install ffmpeg`\n\n# Want to help?\n\nSee the [contribution guidelines](https://github.com/jonatasgrosman/huggingsound/blob/master/CONTRIBUTING.md)\nif you\'d like to contribute to HuggingSound project.\n\nYou don\'t even need to know how to code to contribute to the project. Even the improvement of our documentation is an outstanding contribution.\n\nIf this project has been useful for you, please share it with your friends. This project could be helpful for them too.\n\nIf you like this project and want to motivate the maintainers, give us a :star:. This kind of recognition will make us very happy with the work that we\'ve done with :heart:\n\nYou can also [!["Buy Me A Coffee"](https://www.buymeacoffee.com/assets/img/custom_images/yellow_img.png)](https://www.buymeacoffee.com/jonatasgrosman)\n\n# Citation\nIf you want to cite the tool you can use this:\n\n```bibtex\n@misc{grosman2022huggingsound,\n  title={HuggingSound},\n  author={Grosman, Jonatas},\n  publisher={GitHub},\n  journal={GitHub repository},\n  howpublished={\\url{https://github.com/jonatasgrosman/huggingsound}},\n  year={2022}\n}\n```',
    'author': 'Jonatas Grosman',
    'author_email': 'jonatasgrosman@gmail.com',
    'maintainer': 'Jonatas Grosman',
    'maintainer_email': 'jonatasgrosman@gmail.com',
    'url': 'https://github.com/jonatasgrosman/huggingsound',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)



from __future__ import annotations
import os
import sys
import torch
import logging
import numpy as np
from typing import Any, Optional, Union, Callable
from dataclasses import dataclass, field, asdict
from datasets import Dataset
from transformers import (
    Wav2Vec2Processor, 
    HfArgumentParser, 
    TrainingArguments as HFTrainingArguments, 
    set_seed,
    Trainer,
    AutoConfig,
    AutoModelForCTC,
    EarlyStoppingCallback
)
from transformers.trainer_utils import get_last_checkpoint, is_main_process
from transformers.trainer_pt_utils import get_parameter_names
from huggingsound.metrics import cer, wer

logger = logging.getLogger(__name__)

@dataclass
class ModelArguments:
    """
    Model arguments

    Parameters:
    ----------
    freeze_feature_extractor: Optional[bool] = True
        Whether to freeze the feature extractor layers of the model.

    attention_dropout: Optional[float] = 0.1
        The dropout ratio for the attention probabilities.

    activation_dropout: Optional[float] = 0.1
        The dropout ratio for activations inside the fully connected layer.

    hidden_dropout: Optional[float] = 0.1
        The dropout probabilitiy for all fully connected layers in the embeddings, encoder, and pooler.

    feat_proj_dropout: Optional[float] = 0.1
        The dropout probabilitiy for all 1D convolutional layers in feature extractor.

    final_dropout: Optional[float] = 0.1
        The dropout probabilitiy for CTC head.

    mask_time_prob: Optional[float] = 0.05
        Propability of each feature vector along the time axis to be chosen as the start of the vector
        span to be masked. Approximately ``mask_time_prob * sequence_length // mask_time_length`` feature
        vectors will be masked along the time axis. This is only relevant if ``apply_spec_augment is True``.

    mask_time_length: Optional[int] = 10
        Length of vector span to mask along the time axis.

    mask_feature_prob: Optional[float] = 0.0
        Propability of each feature vector along the feature axis to be chosen as the start of the vector span to
        be masked. Approximately ``mask_time_prob * hidden_size // mask_time_length`` feature vectors will be
        masked along the time axis. This is only relevant if ``apply_spec_augment is True``."

    mask_feature_length: Optional[int] = 10
        Length of vector span to mask along the feature axis.

    layerdrop: Optional[float] = 0.0
        The probability to randomly drops layers at training time. For reference see https://arxiv.org/abs/1909.11556

    apply_spec_augment: Optional[bool] = True
        Whether to apply *SpecAugment* data augmentation to the outputs of the feature extractor. For reference see <https://arxiv.org/abs/1904.08779>"
    
    ctc_loss_reduction: Optional[str] = "sum"
        Specifies the reduction to apply to the output of ``torch.nn.CTCLoss``. Can be 'sum' or 'mean'.
    
    ctc_zero_infinity: Optional[bool] = False
        Whether to zero infinite losses and the associated gradients of ``torch.nn.CTCLoss``. Infinite losses
        mainly occur when the inputs are too short to be aligned to the targets
    """

    freeze_feature_extractor: bool = field(default=True)
    attention_dropout: float = field(default=0.1)
    activation_dropout: float = field(default=0.1)
    hidden_dropout: float = field(default=0.1)
    feat_proj_dropout: float = field(default=0.1)
    final_dropout: float = field(default=0.1)
    mask_time_prob: float = field(default=0.05)
    mask_time_length: int = field(default=10)
    mask_feature_prob: float = field(default=0.0)
    mask_feature_length: int = field(default=10)
    layerdrop: float = field(default=0.0)
    apply_spec_augment: bool = field(default=True)
    ctc_loss_reduction: str = field(default="sum")
    ctc_zero_infinity: bool = field(default=False)


@dataclass
class TrainingArguments:
    """
    Training arguments

    Parameters:
    ----------
    overwrite_output_dir: Optional[bool] = False
        If True, it will overwrite the content of the output directory. 
        If False, it will continue training if output directory points to a checkpoint directory.

    ignore_pretrained_weights: Optional[bool] = False
        If True, it will initializar all the model's weight randonly, i.e., it it will not use the pre-trained model weights during training.
    
    dataloader_num_workers: Optional[int] = 0
        Number of subprocesses to use for data loading (PyTorch only). 0 means that the data will be loaded in the main process.
    
    learning_rate: Optional[float] = 3e-4
        The learning rate for AdamW optimizer.
    
    min_learning_rate: Optional[float] = 0.0
        The minimum learning rate for AdamW optimizer. (only used if lr_warmup_steps or lr_decay_steps are set)
    
    weight_decay: Optional[float] = 0.0
        The weight decay to apply (if not zero) to all layers except all bias and LayerNorm weights in AdamW optimizer.
    
    adam_beta1: Optional[float] = 0.9
        The beta1 hyperparameter for the AdamW optimizer.
    
    adam_beta2: Optional[float] = 0.999
        The beta2 hyperparameter for the AdamW optimizer.
    
    adam_epsilon: Optional[float] = 1e-8
        The epsilon hyperparameter for the AdamW optimizer.
    
    max_grad_norm: Optional[float] = 1.0
        Maximum gradient norm (for gradient clipping).
    
    lr_warmup_steps: Optional[int] = 0
        Number of warmup steps for learning rate scheduler at the beginning of training.
    
    lr_decay_steps: Optional[int] = 0
        Number of decay steps for learning rate scheduler at the end of training.
    
    eval_steps: Optional[int] = None
        Run an evaluation every X steps.
    
    group_by_length: Optional[bool] = True
        Whether or not to group together samples of roughly the same length in the training dataset (to minimize padding applied and be more efficient).
    
    gradient_accumulation_steps: Optional[int] = 1
        Number of updates steps to accumulate before performing a backward/update pass.

    gradient_checkpointing: Optional[bool] = True
        If True, use gradient checkpointing to save memory at the expense of slower backward pass.
    
    pad_to_multiple_of: Optional[int] = None
        If set will pad the sequence to a multiple of the provided value.
        This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute capability >= 7.5 (Volta).

    per_device_train_batch_size: Optional[int] = 8
        The batch size per GPU/TPU core/CPU for training.
    
    per_device_eval_batch_size: Optional[int] = 8
        The batch size per GPU/TPU core/CPU for evaluation.
    
    fp16: Optional[bool] = False
        Whether to use 16-bit (mixed) precision training instead of 32-bit training.

    use_8bit_optimizer: Optional[bool] = False
        Whether to use 8-bit optimizer.
    
    logging_steps: Optional[int] = 100
        Log every X updates steps.
    
    num_train_epochs: Optional[float] = 3.0
        Total number of training epochs to perform (if not an integer, will perform the decimal part percents of
        the last epoch before stopping training).
    
    max_steps: Optional[int] = 0
        If set to a positive number, the total number of training steps to perform. Overrides num_train_epochs.
    
    report_to: Optional[list[str]] = ["none"]
        The list of integrations to report the results and logs to. Supported platforms are "azure_ml",
        "comet_ml", "mlflow", "tensorboard" and "wandb". Use "all" to report to
        all integrations installed, "none" for no integrations.
    
    save_total_limit: Optional[int] = None
        Limit the total amount of checkpoints. Deletes the older checkpoints in the output_dir. Default is unlimited checkpoints
    
    metric_for_best_model: Optional[str] = "cer"
        The metric to use to compare two different models.

    _n_gpu: Optional[int] = 1
        Numbers of gpus to use for training.
    
    seed: Optional[int] = None
        Random seed that will be set at the beginning of training. To ensure reproducibility across runs.

    training_step_callbacks: list[Callable[transformers.Trainer, float]] = []
        A list of callbacks that will be called at the end of each training step.
        The trainer, model and the current loss will be passed to those functions: f(trainer: transformers.Trainer, loss: float)

    batch_creation_callbacks: list[Callable[ list[dict[str, list[int]]] ]] = []
        A list of callbacks that will be called at the end of each batch creation.
        The batch will be passed to those functions: f(batch: list[dict[str, list[int]]])

    evaluation_callbacks: list[Callable[dict, list, list]] = []
        A list of callbacks that will be called at the end of each evaluation step.
        The metrics (WER, CER), references and predictions will be passed to those functions: f(metrics: dict, references: list, predictions: list)

    metrics_batch_size: Optional[int] = None
        Batch size to use for evaluation. When this value is specified, the evaluation function will chunk the data into 
        batches of the specified size and compute the metrics on each batch.
        After all batches are computed, the function will compute the average metrics over all batches.
        (You will probably need to define this if you have memory issues).
    
    show_dataset_stats: Optional[bool] = False
        Whether to show dataset stats

    early_stopping_patience: Optional[int] = None
        Use with metric_for_best_model to stop training when the specified metric worsens for early_stopping_patience evaluation calls.

    load_best_model_at_end: Optional[bool] = False
        Whether to load the best model at the end of training.

    """
    
    overwrite_output_dir: bool = field(default=False)
    ignore_pretrained_weights: bool = field(default=False)
    dataloader_num_workers: int = field(default=0)
    learning_rate: float = field(default=3e-4)
    min_learning_rate: float = field(default=0.0)
    weight_decay: float = field(default=0.0)
    adam_beta1: float = field(default=0.9)
    adam_beta2: float = field(default=0.999)
    adam_epsilon: float = field(default=1e-8)
    max_grad_norm: float = field(default=1.0)
    lr_warmup_steps: int = field(default=0)
    lr_decay_steps: int = field(default=0)
    eval_steps: int = field(default=None)
    group_by_length: bool = field(default=True)
    length_column_name: str = field(default="length")
    gradient_accumulation_steps: int = field(default=1)
    gradient_checkpointing: bool = field(default=True)
    pad_to_multiple_of: int = field(default=None)
    per_device_train_batch_size: int = field(default=8)
    per_device_eval_batch_size: int = field(default=8)
    fp16: bool = field(default=False)
    use_8bit_optimizer: bool = field(default=False)
    logging_steps: int = field(default=100)
    num_train_epochs: float = field(default=3.0)
    max_steps: int = field(default=0)
    report_to: list[str] = field(default_factory=lambda: ["none"])
    save_total_limit: int = field(default=None)
    metric_for_best_model: str = field(default=None)
    _n_gpu: int = field(default=1)
    seed: int = field(default=42)
    training_step_callbacks: list[Callable] = field(default_factory=lambda: [])
    batch_creation_callbacks: list[Callable] = field(default_factory=lambda: [])
    evaluation_callbacks: list[Callable] = field(default_factory=lambda: [])
    metrics_batch_size: int = field(default=None)
    show_dataset_stats: bool = field(default=True)
    early_stopping_patience: int = field(default=None)
    load_best_model_at_end: bool = field(default=False)

class CTCDataCollatorWithPadding:
    """
    Data collator that will dynamically pad the inputs received.

    Parameters:
    ----------
        processor: Wav2Vec2Processor
            The processor used for proccessing the data.

        batch_creation_callbacks: Optional[list[dict[str, Union[list[int], torch.Tensor]]]] = None
            A list of callbacks that will be called at the end of each batch creation.

        pad_to_multiple_of: Optional[int] = None
            If set will pad the sequence to a multiple of the provided value.
            This is especially useful to enable the use of Tensor Cores on NVIDIA hardware with compute capability >= 7.5 (Volta).
    """

    def __init__(self, processor: Wav2Vec2Processor, batch_creation_callbacks: Optional[list[Callable]] = None, 
                 pad_to_multiple_of: Optional[int] = None):
        self.processor = processor
        self.batch_creation_callbacks = batch_creation_callbacks
        self.pad_to_multiple_of = pad_to_multiple_of

    def __call__(self, features: list[dict[str, Union[list[int], torch.Tensor]]]) -> dict[str, torch.Tensor]:
        # split inputs and labels since they have to be of different lenghts and need
        # different padding methods

        if self.batch_creation_callbacks is not None:
            for callback in self.batch_creation_callbacks:
                callback(features)

        input_features = [{"input_values": feature["input_values"]} for feature in features]
        label_features = [{"input_ids": feature["labels"]} for feature in features]

        batch = self.processor.pad(
            input_features,
            padding=True,
            pad_to_multiple_of=self.pad_to_multiple_of,
            return_tensors="pt",
        )
        with self.processor.as_target_processor():
            labels_batch = self.processor.pad(
                label_features,
                padding=True,
                pad_to_multiple_of=self.pad_to_multiple_of,
                return_tensors="pt",
            )

        # replace padding with -100 to ignore loss correctly
        batch["labels"] = labels_batch["input_ids"].masked_fill(labels_batch.attention_mask.ne(1), -100)

        return batch


class CTCTrainer(Trainer):

    """
    Data collator that will dynamically pad the inputs received.

    Parameters:
    ----------
        min_learning_rate: Optional[float] = None
            The minimum learning rate to use.

        lr_warmup_steps: Optional[int] = None
            The number of steps to use for a warmup.

        lr_decay_steps: Optional[int] = None
            The number of steps to use for a decay.
            
        training_step_callbacks: list[Callable[transformers.Trainer, float]] = []
            A list of callbacks that will be called at the end of each training step.
            The trainer, model and the current loss will be passed to those functions: f(trainer: transformers.Trainer, loss: float)
    """

    def __init__(self, min_learning_rate=None, lr_warmup_steps=None, lr_decay_steps=None, 
                 training_step_callbacks: Optional[list[Callable]] = None, use_8bit_optimizer: Optional[bool] = False, **kwargs):

        super().__init__(**kwargs)
        self.min_learning_rate = min_learning_rate
        self.lr_warmup_steps = lr_warmup_steps
        self.lr_decay_steps = lr_decay_steps
        self.training_step_callbacks = training_step_callbacks
        self.use_8bit_optimizer = use_8bit_optimizer

    def create_optimizer(self):

        decay_parameters = get_parameter_names(self.model, [torch.nn.LayerNorm])
        decay_parameters = [name for name in decay_parameters if "bias" not in name]
        optimizer_grouped_parameters = [
            {
                "params": [p for n, p in self.model.named_parameters() if n in decay_parameters],
                "weight_decay": self.args.weight_decay,
            },
            {
                "params": [p for n, p in self.model.named_parameters() if n not in decay_parameters],
                "weight_decay": 0.0,
            },
        ]

        optimizer_class = torch.optim.AdamW

        if self.use_8bit_optimizer:
            
            try:
                import bitsandbytes as bnb
            except ImportError:
                raise ImportError("To use 8bit optimizer please install the bitsandbytes from https://github.com/facebookresearch/bitsandbytes")

            optimizer_class = bnb.optim.AdamW

        self.optimizer = optimizer_class(
            params=optimizer_grouped_parameters,
            lr=self.args.learning_rate,
            betas=(self.args.adam_beta1, self.args.adam_beta2),
            eps=self.args.adam_epsilon,
        )

        return self.optimizer
    
    def create_scheduler(self, num_training_steps: int, optimizer: torch.optim.Optimizer = None):
        """
        Setup the scheduler. This scheduler supports a warmup phase where the learning rate linearly increases 
        at the beginning of the training. And a decay phase where the learning rate decreases over time at the end of the training.
        
        Parameters:
        ----------
            num_training_steps (int): The number of total training steps to do.
        """

        min_lr_ratio = self.min_learning_rate / self.args.learning_rate

        def _lr_lambda(current_step, num_training_steps=num_training_steps, lr_decay_steps=self.lr_decay_steps,
                       lr_warmup_steps=self.lr_warmup_steps, min_lr_ratio=min_lr_ratio):

            if lr_warmup_steps is not None and current_step < lr_warmup_steps: # warmup phase
                return max(
                    min_lr_ratio, float(current_step) / float(max(1, lr_warmup_steps))
                )
            elif lr_decay_steps is not None and current_step >= num_training_steps - lr_decay_steps: # decay phase
                return max(
                    min_lr_ratio, float(num_training_steps - current_step) / float(max(1, lr_decay_steps))
                )
            else: # constant learning rate phase
                return 1
        
        self.lr_scheduler = torch.optim.lr_scheduler.LambdaLR(self.optimizer, _lr_lambda)


    def training_step(self, model: torch.nn.Module, inputs: dict[str, Union[torch.Tensor, Any]]) -> torch.Tensor:
        """
        Perform a training step on a batch of inputs.

        Parameters:
        ----------
            model (:obj:`nn.Module`):
                The model to train.
            inputs (:obj:`dict[str, Union[torch.Tensor, Any]]`):
                The inputs and targets of the model.

                The dictionary will be unpacked before being fed to the model. Most models expect the targets under the
                argument :obj:`labels`. Check your model's documentation for all accepted arguments.

        Returns:
        ----------
            :obj:`torch.Tensor`: The tensor with training loss on this batch.
        """

        model.train()
        inputs = self._prepare_inputs(inputs)

        if self.use_amp:
            with torch.cuda.amp.autocast():
                loss = self.compute_loss(model, inputs)
        else:
            loss = self.compute_loss(model, inputs)

        if self.args.n_gpu > 1:
            if model.module.config.ctc_loss_reduction == "mean":
                loss = loss.mean()
            elif model.module.config.ctc_loss_reduction == "sum":
                loss = loss.sum() / (inputs["labels"] >= 0).sum()
            else:
                raise ValueError(f"{model.config.ctc_loss_reduction} is not valid. Choose one of ['mean', 'sum']")

        if self.args.gradient_accumulation_steps > 1:
            loss = loss / self.args.gradient_accumulation_steps

        if self.use_amp:
            self.scaler.scale(loss).backward()
        elif self.deepspeed:
            self.deepspeed.backward(loss)
        else:
            loss.backward()

        if self.training_step_callbacks is not None:
            for callback in self.training_step_callbacks:
                callback(self, float(loss.detach()))

        return loss.detach()


def finetune_ctc(model_name_or_path: str, output_dir: str, processor: Wav2Vec2Processor, train_dataset: Dataset, 
                 eval_dataset: Optional[Dataset] = None, device: Optional[str] = "cpu", 
                 training_args: Optional[TrainingArguments] = None, model_args: Optional[ModelArguments] = None):
    """
    Finetunes the CTC model.

    Parameters:
    ----------
    
    model_name_or_path : str
        The path to the model or the model identifier from huggingface.co/models.

    output_dir : str
        The output directory where the model checkpoints will be written.

    processor : transformers.Wav2Vec2Processor
        The audio processor object.

    train_dataset : transformers.Dataset
        The training dataset.
    
    eval_dataset : Optional[transformers.Dataset] = None
        The evaluation dataset.
    
    device : str = "cpu"
        The device to run the model training on.

    training_args : Optional[TrainingArguments] = None
        The training arguments.

    model_args : Optional[ModelArguments] = None
        The model arguments.
        
    """

    if training_args is None:
        training_args = TrainingArguments()

    if model_args is None:
        model_args = ModelArguments()

    hf_arg_parser = HfArgumentParser((HFTrainingArguments))
    training_args_dict = asdict(training_args)
    training_args_dict["output_dir"] = output_dir
    training_args_dict["device"] = device
    training_args_dict["no_cuda"] = device == "cpu"
    training_args_dict["do_train"] = True

    if eval_dataset is not None:
        training_args_dict["do_eval"] = True
        training_args_dict["evaluation_strategy"] = "steps"
        training_args_dict["greater_is_better"] = False
    
    hftraining_args = hf_arg_parser.parse_dict(training_args_dict)[0]

    # Set seed before initializing model
    if hftraining_args.seed is not None:
        set_seed(hftraining_args.seed)

    # Detecting last checkpoint.
    last_checkpoint = None
    if os.path.isdir(hftraining_args.output_dir) and not hftraining_args.overwrite_output_dir:
        last_checkpoint = get_last_checkpoint(hftraining_args.output_dir)
        if last_checkpoint is None and len(os.listdir(hftraining_args.output_dir)) > 0:
            raise ValueError(
                f"Output directory ({hftraining_args.output_dir}) already exists and is not empty. "
                "Clear the folder or set TrainingArguments.overwrite_output_dir to True to overcome that."
            )
        elif last_checkpoint is not None:
            logger.info(
                f"Checkpoint detected, resuming training at {last_checkpoint}. To avoid this behavior, change "
                "the model's output directory or set TrainingArguments.overwrite_output_dir to False to train from scratch."
            )

    # Setup logging
    logging.basicConfig(
        format="%(asctime)s - %(levelname)s - %(name)s - %(message)s",
        datefmt="%m/%d/%Y %H:%M:%S",
        handlers=[logging.StreamHandler(sys.stdout)],
    )
    logger.setLevel(logging.INFO if is_main_process(hftraining_args.local_rank) else logging.WARN)

    if training_args.show_dataset_stats:
        logger.info("Getting dataset stats...")

        total_train_size_in_seconds = 0
        for sample in train_dataset:
            total_train_size_in_seconds += len(sample["input_values"]) / processor.feature_extractor.sampling_rate
        logger.info(f"Training dataset size: {len(train_dataset)} samples, {total_train_size_in_seconds/3600} hours")

        if eval_dataset is not None:
            total_eval_size_in_seconds = 0
            for sample in eval_dataset:
                total_eval_size_in_seconds += len(sample["input_values"]) / processor.feature_extractor.sampling_rate
            logger.info(f"Evaluation dataset size: {len(eval_dataset)} samples, {total_eval_size_in_seconds/3600} hours")

    # Setup model

    config = AutoConfig.from_pretrained(model_name_or_path)

    config.update(
        {
            "feat_proj_dropout": model_args.feat_proj_dropout,
            "attention_dropout": model_args.attention_dropout,
            "hidden_dropout": model_args.hidden_dropout,
            "final_dropout": model_args.final_dropout,
            "mask_time_prob": model_args.mask_time_prob,
            "mask_time_length": model_args.mask_time_length,
            "mask_feature_prob": model_args.mask_feature_prob,
            "mask_feature_length": model_args.mask_feature_length,
            "layerdrop": model_args.layerdrop,
            "activation_dropout": model_args.activation_dropout,
            "ctc_loss_reduction": model_args.ctc_loss_reduction,
            "pad_token_id": processor.tokenizer.pad_token_id,
            "vocab_size": len(processor.tokenizer),
        }
    )

    if training_args.ignore_pretrained_weights:
        model = AutoModelForCTC.from_config(config=config)
    else:
        model = AutoModelForCTC.from_pretrained(model_name_or_path, config=config)

    if hftraining_args.gradient_checkpointing:
        model.gradient_checkpointing_enable()

    if model_args.freeze_feature_extractor:    
        model.freeze_feature_extractor()

    # Defining metrics
    def _compute_metrics(pred, processor=processor, metrics_batch_size=training_args.metrics_batch_size, 
                         evaluation_callbacks=training_args.evaluation_callbacks):

        pred_logits = pred.predictions
        pred_ids = np.argmax(pred_logits, axis=-1)

        pred.label_ids[pred.label_ids == -100] = processor.tokenizer.pad_token_id

        pred_str = processor.batch_decode(pred_ids)
        
        # we do not want to group tokens when computing the metrics
        label_str = processor.batch_decode(pred.label_ids, group_tokens=False)

        metrics = {
            "wer": wer(predictions=pred_str, references=label_str, chunk_size=metrics_batch_size),
            "cer": cer(predictions=pred_str, references=label_str, chunk_size=metrics_batch_size),
        }

        if evaluation_callbacks is not None:
            for callback in evaluation_callbacks:
                callback(metrics, label_str, pred_str)

        return metrics

    # Setup data collator

    data_collator = CTCDataCollatorWithPadding(
        processor=processor,
        batch_creation_callbacks=training_args.batch_creation_callbacks,
        pad_to_multiple_of=training_args.pad_to_multiple_of,
    )
    
    # Initialize Trainer

    logger.info("Building trainer...")

    callbacks = []
    if training_args.early_stopping_patience is not None and training_args.early_stopping_patience > 0:
        callbacks.append(
            EarlyStoppingCallback(early_stopping_patience=training_args.early_stopping_patience)
        )
        hftraining_args.load_best_model_at_end = True

    trainer = CTCTrainer(
        args=hftraining_args,
        model = model,
        data_collator=data_collator,
        train_dataset=train_dataset,
        eval_dataset=eval_dataset,
        compute_metrics=_compute_metrics if eval_dataset is not None else None,
        min_learning_rate=training_args.min_learning_rate,
        lr_warmup_steps=training_args.lr_warmup_steps,
        lr_decay_steps=training_args.lr_decay_steps,
        training_step_callbacks=training_args.training_step_callbacks,
        use_8bit_optimizer=training_args.use_8bit_optimizer,
        callbacks = callbacks
    )

    # Training

    logger.info("Starting training...")

    if last_checkpoint is not None:
        checkpoint = last_checkpoint
    elif os.path.isdir(model_name_or_path):
        checkpoint = model_name_or_path
    else:
        checkpoint = None

    # Saving processor config to output directory

    if is_main_process(hftraining_args.local_rank):
        processor.save_pretrained(hftraining_args.output_dir)
    
    train_result = trainer.train(resume_from_checkpoint=checkpoint)
    trainer.save_model()

    metrics = train_result.metrics  
    metrics["train_samples"] = len(train_dataset)

    trainer.log_metrics("train", metrics)
    trainer.save_metrics("train", metrics)
    trainer.save_state()
